var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
import { isOutside, getInsidePosition } from "./Coordinate";
import { map, equal } from "./utils";
var InputObserver = /** @class */ (function () {
    function InputObserver(_a) {
        var options = _a.options, itm = _a.itm, em = _a.em, axm = _a.axm, am = _a.am;
        this.isOutside = false;
        this.moveDistance = null;
        this.isStopped = false;
        this.options = options;
        this.itm = itm;
        this.em = em;
        this.axm = axm;
        this.am = am;
    }
    // when move pointer is held in outside
    InputObserver.prototype.atOutside = function (pos) {
        var _this = this;
        if (this.isOutside) {
            return this.axm.map(pos, function (v, opt) {
                var tn = opt.range[0] - opt.bounce[0];
                var tx = opt.range[1] + opt.bounce[1];
                return v > tx ? tx : (v < tn ? tn : v);
            });
        }
        else {
            // when start pointer is held in inside
            // get a initialization slope value to prevent smooth animation.
            var initSlope_1 = this.am.easing(0.00001) / 0.00001;
            return this.axm.map(pos, function (v, opt) {
                var min = opt.range[0];
                var max = opt.range[1];
                var out = opt.bounce;
                var circular = opt.circular;
                if (circular && (circular[0] || circular[1])) {
                    return v;
                }
                else if (v < min) { // left
                    return min - _this.am.easing((min - v) / (out[0] * initSlope_1)) * out[0];
                }
                else if (v > max) { // right
                    return max + _this.am.easing((v - max) / (out[1] * initSlope_1)) * out[1];
                }
                return v;
            });
        }
    };
    InputObserver.prototype.get = function (input) {
        return this.axm.get(input.axes);
    };
    InputObserver.prototype.hold = function (input, event) {
        if (this.itm.isInterrupted() || !input.axes.length) {
            return;
        }
        var changeOption = {
            input: input,
            event: event,
        };
        this.isStopped = false;
        this.itm.setInterrupt(true);
        this.am.grab(input.axes, changeOption);
        !this.moveDistance && this.em.triggerHold(this.axm.get(), changeOption);
        this.isOutside = this.axm.isOutside(input.axes);
        this.moveDistance = this.axm.get(input.axes);
    };
    InputObserver.prototype.change = function (input, event, offset) {
        if (this.isStopped || !this.itm.isInterrupting() || this.axm.every(offset, function (v) { return v === 0; })) {
            return;
        }
        var depaPos = this.moveDistance || this.axm.get(input.axes);
        var destPos;
        // for outside logic
        destPos = map(depaPos, function (v, k) { return v + (offset[k] || 0); });
        this.moveDistance && (this.moveDistance = destPos);
        // from outside to inside
        if (this.isOutside &&
            this.axm.every(depaPos, function (v, opt) { return !isOutside(v, opt.range); })) {
            this.isOutside = false;
        }
        destPos = this.atOutside(destPos);
        var isCanceled = !this.em.triggerChange(destPos, false, depaPos, {
            input: input,
            event: event,
        }, true);
        if (isCanceled) {
            this.isStopped = true;
            this.moveDistance = null;
            this.am.finish(false);
        }
    };
    InputObserver.prototype.release = function (input, event, offset, inputDuration) {
        if (this.isStopped || !this.itm.isInterrupting() || !this.moveDistance) {
            return;
        }
        var pos = this.axm.get(input.axes);
        var depaPos = this.axm.get();
        var destPos = this.axm.get(this.axm.map(offset, function (v, opt, k) {
            if (opt.circular && (opt.circular[0] || opt.circular[1])) {
                return pos[k] + v;
            }
            else {
                return getInsidePosition(pos[k] + v, opt.range, opt.circular, opt.bounce);
            }
        }));
        var duration = this.am.getDuration(destPos, pos, inputDuration);
        if (duration === 0) {
            destPos = __assign({}, depaPos);
        }
        // prepare params
        var param = {
            depaPos: depaPos,
            destPos: destPos,
            duration: duration,
            delta: this.axm.getDelta(depaPos, destPos),
            inputEvent: event,
            input: input,
            isTrusted: true,
        };
        this.em.triggerRelease(param);
        this.moveDistance = null;
        // to contol
        var userWish = this.am.getUserControll(param);
        var isEqual = equal(userWish.destPos, depaPos);
        var changeOption = {
            input: input,
            event: event,
        };
        if (isEqual || userWish.duration === 0) {
            !isEqual && this.em.triggerChange(userWish.destPos, false, depaPos, changeOption, true);
            this.itm.setInterrupt(false);
            if (this.axm.isOutside()) {
                this.am.restore(changeOption);
            }
            else {
                this.em.triggerFinish(true);
            }
        }
        else {
            this.am.animateTo(userWish.destPos, userWish.duration, changeOption);
        }
    };
    return InputObserver;
}());
export { InputObserver };
//# sourceMappingURL=InputObserver.js.map